"""
AWS test utils
"""

import boto3


def ssm_put_parameter_securestring(name="test", value="abc"):
    """
    Create mocked SSM parameter to later fetch using moto
    """
    ssm = boto3.client("ssm")
    ssm.put_parameter(
        Name=name,
        Value=value,
        Type="SecureString",
    )


def put_secretsmanager_secret(name="test", value=None, binary=None):
    """
    Create mocked Secrets Manager secret to later fetch using moto
    """
    ssm = boto3.client("secretsmanager")
    if value is not None:
        ssm.create_secret(Name=name, SecretString=value)

    if binary is not None:
        ssm.create_secret(Name=name, SecretBinary=binary)



"""
Base using environment variables unit tests
"""

import pytest
from value_fetcher import ValueFetcher

fetcher = ValueFetcher()


def test_env_vars(monkeypatch):
    """
    Test fetching config values from environment
    """

    # Test fetches from environment by default
    monkeypatch.setenv("ENV_VAR_TEST", "test_value")
    assert fetcher.get("ENV_VAR_TEST") == "test_value"

    # Test non-existent variable throws an exception
    with pytest.raises(ValueError) as exception:
        fetcher.get("NONEXISTENT_ENV_VAR")

    assert "Missing or empty environment value" in str(exception.value)


def test_env_vars_with_defaults():
    """
    Test fetching environment variables with defaults set
    """

    fetcher_with_defaults = ValueFetcher({"SET_WITH_DEFAULT": "some_default"})

    assert fetcher_with_defaults.get("SET_WITH_DEFAULT") == "some_default"



"""
Pytest unit tests for aws client calls
"""

import os
from moto import mock_aws
from value_fetcher.aws import Aws
from tests import utils

# Set boto client default values
os.environ["AWS_DEFAULT_REGION"] = "eu-west-2"


@mock_aws
def test_getting_parameter():
    """
    Check an SSM parameter is returned
    """

    # Create test parameter to later fetch using moto
    utils.ssm_put_parameter_securestring()

    # Fetch and verify parameter
    aws = Aws()
    param = aws.get_parameter_value("test")
    assert param == "abc"

    # Assert missing parameter returns nothing
    assert not aws.get_parameter_value("doesnotexist")


@mock_aws
def test_getting_secret():
    """
    Check string and binary secrets may be retrieved
    """

    # Create mocked secrets to later fetch using moto
    secret_string = "my secret value"
    secret_binary = bytes(secret_string, "utf-8")
    utils.put_secretsmanager_secret("/secret/string", secret_string)
    utils.put_secretsmanager_secret("/secret/binary", None, secret_binary)

    # Fetch secrets
    aws = Aws()
    assert aws.get_secret_value("/secret/string") == secret_string
    assert aws.get_secret_value("/secret/binary") == secret_string
    assert not aws.get_secret_value("/does/not/exist")



"""
Provider unit tests
"""

import os

import pytest
from value_fetcher import ValueFetcher

# Set mocked boto (moto) client default values
os.environ["AWS_DEFAULT_REGION"] = "eu-west-2"

fetcher = ValueFetcher()


def test_invalid_names():
    """
    Assert invalid or empty names produce exceptions
    """
    # Test non-existent variable throws an exception
    with pytest.raises(ValueError) as exception:
        fetcher.get("")

    with pytest.raises(ValueError) as env_exception:
        fetcher.get_from_env("")

    assert str(exception.value) == "Invalid key name provided"
    assert str(env_exception.value) == "Missing or empty environment key"


def test_unknown_source(monkeypatch):
    """
    Test unknown source raises an exception
    """
    monkeypatch.setenv("UNKNOWN", "test_value")
    monkeypatch.setenv("UNKNOWN_SOURCE", "unknown")
    # Test unknown source returns nothing
    with pytest.raises(ValueError) as exception:
        fetcher.get("UNKNOWN")

    assert "Unknown source" in str(exception.value)



"""
Base using AWS Systems Manager Parameter Store unit tests
"""

import os

import pytest
from moto import mock_aws
from tests import utils
from value_fetcher import ValueFetcher

# Set mocked boto (moto) client default values
os.environ["AWS_DEFAULT_REGION"] = "eu-west-2"

fetcher = ValueFetcher()


@mock_aws
def test_aws_ssm_parameter_store_called_direct_no_key_raises_error():
    """
    Test fetching from parameter store without specifying a key raises
    exception
    """

    # Assert an exception is thrown if parameter key is not provided
    with pytest.raises(ValueError) as exception:
        fetcher.get_from_aws_ssm_parameter_store(None)

    val = exception.value
    assert "Missing or empty AWS SSM Parameter Store key" in str(val)

    # Assert an exception is thrown if parameter key is empty
    with pytest.raises(ValueError) as exception:
        fetcher.get_from_aws_ssm_parameter_store("")

    val = exception.value
    assert "Missing or empty AWS SSM Parameter Store key" in str(val)


@mock_aws
def test_aws_ssm_parameter_store_with_env_var_config_success(monkeypatch):
    """
    Test fetching config values from parameter store with key name from
    environment
    """

    # Instruct config provider to fetch value for SSM_TEST from parameter store
    monkeypatch.setenv("SSM_OK_SOURCE", "aws_ssm_parameter_store")

    # Create mocked test parameter to later fetch using moto
    utils.ssm_put_parameter_securestring("/is/ok", "my val")

    # Assert value is correctly retrieved if the key name is provided
    monkeypatch.setenv("SSM_OK_PARAMETER_STORE_NAME", "/is/ok")
    assert fetcher.get("SSM_OK") == "my val"


@mock_aws
def test_aws_ssm_parameter_store_without_env_var_config_success():
    """
    Test fetching config values from parameter store with no additional
    environment variables.
    """

    key = "/is/ok/no/env/vars"
    value = "my val no env var"

    # Create mocked test parameter to later fetch using moto
    utils.ssm_put_parameter_securestring(key, value)

    assert fetcher.get_from_aws_ssm_parameter_store(key) == value


@mock_aws
def test_aws_ssm_parameter_store_does_not_exist(monkeypatch):
    """
    Test fetching non-existent config values from parameter store
    """
    key = "SSM_NONEXISTENT_SOURCE"
    monkeypatch.setenv(key, "aws_ssm_parameter_store")

    key = "SSM_NONEXISTENT_PARAMETER_STORE_NAME"
    monkeypatch.setenv(key, "/does/not/exist")

    # Assert an exception is thrown if parameter name is not provided
    with pytest.raises(ValueError) as exception:
        fetcher.get("SSM_NONEXISTENT")

    assert "Missing or empty AWS SSM Parameter Store" in str(exception.value)



"""
Base using AWS Secrets Manager unit tests
"""

import os

import pytest
from moto import mock_aws
from tests import utils
from value_fetcher import ValueFetcher

# Set mocked boto (moto) client default values
os.environ["AWS_DEFAULT_REGION"] = "eu-west-2"

fetcher = ValueFetcher()


@mock_aws
def test_aws_secrets_manager_called_direct_no_key_raises_error():
    """
    Test fetching from secrets manager without specifying a key raises
    exception
    """

    # Assert an exception is thrown if parameter key is not provided
    with pytest.raises(ValueError) as exception:
        fetcher.get_from_aws_secrets_manager(None)

    assert "Missing or empty AWS Secrets Manager key" in str(exception.value)

    # Assert an exception is thrown if parameter key is empty
    with pytest.raises(ValueError) as exception:
        fetcher.get_from_aws_secrets_manager("")

    assert "Missing or empty AWS Secrets Manager key" in str(exception.value)


@mock_aws
def test_aws_secrets_manager_with_env_var_config_success(monkeypatch):
    """
    Test fetching config values from secrets manager
    """

    # Instruct config provider to fetch value for SM_TEST from secrets manager
    monkeypatch.setenv("SECRET_OK_SOURCE", "aws_secrets_manager")

    # Create mocked test parameter to later fetch using moto
    utils.put_secretsmanager_secret("/is/ok", "my secret")

    # Assert value is correctly retrieved if the key name is provided
    monkeypatch.setenv("SECRET_OK_SECRETS_MANAGER_NAME", "/is/ok")
    assert fetcher.get("SECRET_OK") == "my secret"


@mock_aws
def test_aws_secrets_manager_without_env_var_config_success():
    """
    Test fetching config values from secrets manager without specifying a name
    """

    key = "/is/ok/no/env/vars"
    value = "my secret val no env var"

    # Create mocked test parameter to later fetch using moto
    utils.put_secretsmanager_secret(key, value)

    assert fetcher.get_from_aws_secrets_manager(key) == value


@mock_aws
def test_aws_secrets_manager_does_not_exist(monkeypatch):
    """
    Test fetching non-existent config values from secrets manager
    """

    key = "SECRET_NONEXISTENT_SOURCE"
    monkeypatch.setenv(key, "aws_secrets_manager")

    key = "SECRET_NONEXISTENT_SECRETS_MANAGER_NAME"
    monkeypatch.setenv(key, "/does/not/exist")

    # Assert an exception is thrown if secret name is not provided
    with pytest.raises(ValueError) as exception:
        fetcher.get("SECRET_NONEXISTENT")

    msg = "Missing or empty AWS Secrets Manager value for /does/not/exist"
    assert str(exception.value) == msg



"""
Configuration provider to fetch values from the following sources:
- Environment variable
- AWS Systems Manager (SSM) Parameter Store
- AWS Secrets Manager
"""

import os
import logging
from .aws import Aws


class ValueFetcher:
    """
    Configuration provider class
    """

    def __init__(self, env_defaults: dict = None) -> None:
        self.env_defaults = env_defaults

    def get(self, key: str) -> str:
        """
        Given a key name MY_VAL, check for a source env var e.g. MY_VAL_SOURCE.
        Raises exception if source does not match list of known services.
        """
        # Make the provided name all-uppercase
        key = key.upper()

        if not isinstance(key, str) or len(key) == 0:
            message = "Invalid key name provided"
            logging.critical(message)
            logging.critical(key)
            raise ValueError(message)

        logging.debug("Retrieving value for key %s", key)
        source = os.environ.get(f"{key}_SOURCE", "env").lower()
        logging.debug("%s source: %s", key, source)
        # Get value from environment variable
        if source == "env":
            return self.get_from_env(key)

        # Get value from AWS SSM parameter store
        if source == "aws_ssm_parameter_store":
            return self.get_from_aws_ssm_parameter_store(key)

        # Get value from AWS Secrets Manager
        if source == "aws_secrets_manager":
            return self.get_from_aws_secrets_manager(key)

        message = f"Unknown source {source}"
        logging.critical(message)
        raise ValueError(message)

    def get_from_env(self, key) -> str:
        """
        Fetch a value from environment variables, using a default if available.
        If value is empty and cannot be found in defaults, raise exception.
        """

        if not isinstance(key, str) or len(key) == 0:
            message = "Missing or empty environment key"
            logging.critical(message)
            logging.critical(key)
            raise ValueError(message)

        logging.debug("Checking environment for key: %s", key)
        value = os.environ.get(key, "").strip()
        if not isinstance(value, str) or len(value) == 0:
            defaults = self.env_defaults
            if isinstance(defaults, dict) and key in defaults:
                logging.debug("Using default value for env var: %s", key)
                value = defaults[key]
            else:
                message = f"Missing or empty environment value for {key}"
                logging.critical(message)
                raise ValueError(message)

        return value

    def get_from_aws_ssm_parameter_store(self, key: str) -> str:
        """
        Fetch a value from AWS SSM Parameter store.
        If the parameter name is not configured by an environment variable,
        try the provided key.
        """

        if not isinstance(key, str) or len(key) == 0:
            message = "Missing or empty AWS SSM Parameter Store key"
            logging.critical(message)
            raise ValueError(message)

        name_key = f"{key}_PARAMETER_STORE_NAME"

        logging.debug("Checking env for AWS SSM Parameter %s", name_key)
        try:
            name = self.get_from_env(name_key)
            logging.debug("Using parameter name %s", name)
        except ValueError as exception:
            logging.debug(exception)
            logging.debug("Using key name for parameter %s", key)
            name = key

        aws = Aws()
        value = aws.get_parameter_value(name)
        if not isinstance(value, str) or len(value) == 0:
            message = f"Missing or empty AWS SSM Parameter Store {name}"
            logging.critical(message)
            raise ValueError(message)

        return value

    def get_from_aws_secrets_manager(self, key: str) -> str:
        """
        Fetch a value from AWS Secrets Manager.
        If the secret name is not configured by an environment variable,
        use the key directly.
        """

        if not isinstance(key, str) or len(key) == 0:
            message = "Missing or empty AWS Secrets Manager key"
            logging.critical(message)
            raise ValueError(message)

        name_key = f"{key}_SECRETS_MANAGER_NAME"

        logging.debug("Checking env for AWS Secret Manager %s", name_key)
        try:
            name = self.get_from_env(name_key)
            logging.debug("Using secret name %s", name)
        except ValueError as exception:
            logging.debug(exception)
            logging.debug("Using key name for secret %s", key)
            name = key

        aws = Aws()
        value = aws.get_secret_value(name)
        if not isinstance(value, str) or len(value) == 0:
            message = f"Missing or empty AWS Secrets Manager value for {name}"
            logging.critical(message)
            raise ValueError(message)

        return value
